module PC.UActions      exposing ( uCodes
                                 , uAction2Label
                                 , string2uAction
                                 , uAction2UCode )

import Array            exposing (Array)
import Tuple

import PC.Types         exposing (..)
import PC.Helpers       exposing (..)

uCodes : List UCode
uCodes =
    [ UCode UA_Nothing                              "n"         (\s -> s)                       "Empty"                  
    , UCode UA_ALU_Accumulator2DataBus              "acc2db"    actAccumulator2DataBus          "Acc -> DataBus"         
    , UCode UA_ALU_AccumulatorDecrement             "accDec"    actAccumulatorDecrement         "Acc --"                 
    , UCode UA_ALU_AccumulatorIncrement             "accInc"    actAccumulatorIncrement         "Acc ++"                 
    , UCode UA_ALU_AccumulatorShiftLeftByDataBus    "accLsN"    actAccumulatorShiftLeft         "Acc << DataBus"         
    , UCode UA_ALU_AccumulatorShiftRightByDataBus   "accRsN"    actAccumulatorShiftRight        "Acc >> DataBus"         
    , UCode UA_ALU_AccumulatorAddByDataBus          "accAdd"    actAccumulatorAdd               "Acc + DataBus"          
    , UCode UA_ALU_AccumulatorSubtractByDataBus     "accSub"    actAccumulatorSub               "Acc - DataBus"          
    , UCode UA_ALU_AccumulatorMultiplyByDataBus     "accMul"    actAccumulatorMultiply          "Acc * DataBus"          
    , UCode UA_ALU_AccumulatorDivideByDataBus       "accDiv"    actAccumulatorDivide            "Acc / DataBus"          
    , UCode UA_CU_InstructionReg2AddressBus         "ir2ab"     actInstructionReg2AddressBus    "InstReg -> AddrBus"     
    , UCode UA_CU_InstructionReg2DataBus            "ir2db"     actInstructionReg2DataBus       "InstReg -> DataBus"     
    , UCode UA_CU_InstructionReg2ProgrammCounter    "ir2pc"     actInstructionReg2ProgrammCounter "InstReg -> ProgCount" 
    , UCode UA_CU_InstructionReg2UCounter           "ir2uc"     actInstructionReg2UCounter      "InstReg -> µCounter"    
    , UCode UA_CU_ProgrammCounterIncrement          "pcInc"     actProgrammCounterIncrement     "ProgCounter ++"         
    , UCode UA_CU_ProgrammCounter2AddressBus        "pc2ab"     actProgrammCounter2AddressBus   "ProgCounter -> AddrBus" 
    , UCode UA_CU_ResetUCounter                     "ucReset"   actResetUCounter                "µCounter = 0"           
    , UCode UA_CU_InstructionReg2ProgrammCounterIfAccEq0 
                        "ir2pciacceq0" actInstructionReg2ProgrammCounterIfAccEq0 "Acc == 0 => InstReg -> ProgCounter"    
    , UCode UA_DataBus2Accumulator                  "db2acc"    actDataBus2Accumulator          "DataBus -> Acc"         
    , UCode UA_DataBus2InstructionReg               "db2ir"     actDataBus2InstructionReg       "DataBus -> InstReg"     
    , UCode UA_DataBus2Ram                          "db2ram"    actDataBus2Ram                  "DataBus -> Ram"         
    , UCode UA_Ram2DataBus                          "ram2db"    actRam2DataBus                  "Ram -> DataBus"         
    , UCode UA_AlertUser                            "alert"     actDoAlert                      "Signal End of Programm" 
    ]


-- #######################################################################################
-- ACTIONS
-- #######################################################################################

actRam2DataBus : PC -> PC
actRam2DataBus pc =
    let
        ab =
            pc.addressBus

        db =
            Tuple.first <| valueAtRam ab pc.ram
    in
    { pc | dataBus = db }


actDataBus2InstructionReg : PC -> PC
actDataBus2InstructionReg pc =
    { pc | instructionReg = pc.dataBus }


actResetUCounter : PC -> PC
actResetUCounter pc =
    { pc | uCounter = 0 }


actInstructionReg2UCounter : PC -> PC
actInstructionReg2UCounter pc =
    -- Multiply by ten, because every instruction is 10 uCodes long
    let
        (instruction, address) = seperateInstructionsEntry pc.instructionReg
    in
    { pc | uCounter = instruction * 10}


actInstructionReg2AddressBus : PC -> PC
actInstructionReg2AddressBus pc =
    let
        (instruction, address) = seperateInstructionsEntry pc.instructionReg
    in
    { pc | addressBus = address }

actInstructionReg2DataBus : PC -> PC
actInstructionReg2DataBus pc =
    let
        (_, data) = seperateInstructionsEntry pc.instructionReg
    in
    { pc | dataBus = data }

actInstructionReg2ProgrammCounter : PC -> PC
actInstructionReg2ProgrammCounter pc =
    let
        (instruction, address) = seperateInstructionsEntry pc.instructionReg
    in
    
    { pc | programmCounter = address }


actProgrammCounterIncrement : PC -> PC
actProgrammCounterIncrement pc =
    { pc | programmCounter = pc.programmCounter + 1 }


actProgrammCounter2AddressBus : PC -> PC
actProgrammCounter2AddressBus pc =
    { pc | addressBus = pc.programmCounter }

actInstructionReg2ProgrammCounterIfAccEq0 : PC -> PC
actInstructionReg2ProgrammCounterIfAccEq0 pc =
    if pc.accumulator == 0 then
        let
            (_,addr) = seperateInstructionsEntry pc.instructionReg
        in
        { pc | programmCounter = addr }
    else
        pc

actDataBus2Accumulator : PC -> PC
actDataBus2Accumulator pc =
    { pc | accumulator = pc.dataBus }


actAccumulator2DataBus : PC -> PC
actAccumulator2DataBus pc =
    { pc | dataBus = pc.accumulator }


actAccumulatorIncrement : PC -> PC
actAccumulatorIncrement pc =
    { pc | accumulator = pc.accumulator + 1 }


actAccumulatorDecrement : PC -> PC
actAccumulatorDecrement pc =
    { pc | accumulator = pc.accumulator - 1 }

actAccumulatorShiftLeft : PC -> PC
actAccumulatorShiftLeft pc = 
    { pc | accumulator = pc.accumulator * (2 ^ pc.dataBus) }

actAccumulatorShiftRight : PC -> PC
actAccumulatorShiftRight pc = 
    { pc | accumulator = pc.accumulator // (2 ^ pc.dataBus) }

actAccumulatorAdd : PC -> PC
actAccumulatorAdd pc = 
    { pc | accumulator = pc.accumulator + pc.dataBus }

actAccumulatorSub : PC -> PC
actAccumulatorSub pc = 
    { pc | accumulator = pc.accumulator - pc.dataBus }

actAccumulatorMultiply : PC -> PC
actAccumulatorMultiply pc = 
    { pc | accumulator = pc.accumulator * pc.dataBus }

actAccumulatorDivide : PC -> PC
actAccumulatorDivide pc = 
    { pc | accumulator = pc.accumulator // pc.dataBus }



actDataBus2Ram : PC -> PC
actDataBus2Ram pc =
    let
        newRam =
            changeAtRam pc.addressBus pc.dataBus pc.ram
    in
    { pc | ram = newRam }

actDoAlert : PC -> PC
actDoAlert pc = pc      -- This gets caputred by update anyway


-- #################################################################
-- Helpers
-- #################################################################

uAction2Label : UAction -> String
uAction2Label variant =
    let
        filtered_list = List.filter (\u -> u.variant == variant ) uCodes
    in
    case List.head filtered_list of
        Just uCode -> uCode.label
        Nothing -> "IDK"

uAction2UCode : UAction -> UCode
uAction2UCode variant =
    let
        filtered_list = List.filter (\u -> u.variant == variant ) uCodes
    in
    case List.head filtered_list of
        Just uCode  -> uCode
        Nothing     -> UCode UA_Nothing "n" (\s -> s) "Empty" 

string2uAction : String -> Maybe UAction
string2uAction str =
    let
        filtered_list = List.filter (\u -> u.id == str) uCodes
    in
    case List.head filtered_list of
        Just uCode -> Just uCode.variant
        _ -> Nothing

